#define _WIN32_WINDOWS 0x0444
#include <windows.h>
#include <time.h>
#include "zmouse.h"
#include "czerwony_kapturek_jbwheel_WheelTool.h"

//handle of this dll module
HANDLE hm;
//handle to debugging console
HANDLE console = NULL;
//temporary handle to console
HANDLE tmpConsole = NULL;
//hook procedure for posted messages
HHOOK hhPost;
//hook procedure for sent
HHOOK hhSend;
//ID of JBuilder process
DWORD myProcess;

//message for Win95
WORD uMSH_MOUSEWHEEL = 0;

//jbWheel opentool object instance
jobject jo;
//JNI enviroment pointer
JNIEnv * env = NULL;
//pointer to Java VM
JavaVM* vm;
//id of the java method to call
jmethodID mid;

//print debugging messages
void trace( char* msg );

//main dll function
BOOL APIENTRY
DllMain( HANDLE hModule,
         DWORD  ul_reason_for_call,
         LPVOID /*lpReserved*/
        )
{
    hm = hModule;
    switch (ul_reason_for_call)
    {
        case DLL_PROCESS_ATTACH:
        case DLL_THREAD_ATTACH:
        case DLL_THREAD_DETACH:
            break;

        case DLL_PROCESS_DETACH:
            if (tmpConsole)
            {
                FreeConsole();
            }
            if (hhPost)
            {
                ::UnhookWindowsHookEx( hhPost );
                hhPost = 0;
            }
            if (hhSend)
            {
                ::UnhookWindowsHookEx( hhSend );
                hhSend = 0;
            }
            break;
    }
    return TRUE;
}


//process messagess intercepted by hooks
void _stdcall
processHookedMessage(WORD message, HWND hwnd, long lParam, long wParam, bool sent)
{
//    long t = GetTickCount();
    DWORD process;
    GetWindowThreadProcessId( hwnd, &process );
    if ( process == myProcess && ( message == WM_MOUSEWHEEL
        || message == uMSH_MOUSEWHEEL ) )
    {
        bool mw = message == WM_MOUSEWHEEL;
        if (console)
        {
            char buff[50];
            wsprintf(buff, "%s - %s",
                    mw ? "WM_MOUSEWHEEL" : "MSH_MOUSEWHEEL",
                    sent ? "sent" : "posted" );
            trace(buff);
        }
        if (!env)
        {
            vm->AttachCurrentThread( (void**)&env, NULL );
        //}
            // I'm not sure if storing method ID is valid
            jclass jc = env->GetObjectClass( jo );
            mid = env->GetMethodID( jc, "onWheelMoved", "(IIII)V");
        }

        short delta;
        short keys = 0;
        if ( mw )
        {
            delta = (short)HIWORD(wParam);
            keys = (short)LOWORD(wParam);
        }
        else
        {
            if ( GetKeyState( VK_SHIFT )&0xFF00 )
                keys |= MK_SHIFT;
            if ( GetKeyState( VK_CONTROL )&0xFF00 )
                keys |= MK_CONTROL;
            delta = (short)wParam;
        }

        env->CallVoidMethod( jo, mid, delta,
                    (short)LOWORD(lParam),
                    (short)HIWORD(lParam),
                    keys /*+ (alt ? 0x8000 : 0)*/ );

//        if (console)
//        {
//            t = GetTickCount() - t;
//            char buff[255];
//            wsprintf(buff, "processing message took %lu ms", t);
//            trace(buff);
//        }
    }
}


//hook for messages posted to the window - for "normal" mouse drivers
long _stdcall
postHookProc( int code, unsigned int wParam, long lParam )
{
    if (code >= 0)
    {
        MSG* pMsg = (MSG*)lParam;
        processHookedMessage(pMsg->message, pMsg->hwnd, pMsg->lParam, pMsg->wParam, false);
    }
    long ret = CallNextHookEx( hhPost, code, wParam, lParam );
    return ret;
}


//hook for messages sent to the window - needed for IBM mice
long _stdcall
sendHookProc( int code, unsigned int wParam, long lParam )
{
    if (code >= 0)
    {
        CWPSTRUCT* pMsg = (CWPSTRUCT*)lParam;
        processHookedMessage(pMsg->message, pMsg->hwnd, pMsg->lParam, pMsg->wParam, true);
    }
    long ret = CallNextHookEx( hhSend, code, wParam, lParam );
    return ret;
}


/*
 * Class:     czerwony_kapturek_jbwheel_WheelTool
 * Method:    startCaptureWheelMessages
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_czerwony_kapturek_jbwheel_WheelTool_startCaptureWheelMessages(JNIEnv * env1, jobject jo1)
{
    uMSH_MOUSEWHEEL = RegisterWindowMessage(MSH_MOUSEWHEEL);

    jo = env1->NewGlobalRef( jo1 );
    env1->GetJavaVM( &vm );

    hhPost = ::SetWindowsHookEx( WH_GETMESSAGE, postHookProc, (HINSTANCE)hm, NULL );
    if (!hhPost)
    {
        char	buff[255];
        wsprintf( buff, "Failed to register message hook for posted messagess (%08X)", GetLastError() );
        trace( buff );
    }
    hhSend = ::SetWindowsHookEx( WH_CALLWNDPROC, sendHookProc, (HINSTANCE)hm, NULL );
    if (!hhSend)
    {
        char	buff[255];
        wsprintf( buff, "Failed to register message hook for sent messagess (%08X)", GetLastError() );
        trace( buff );
    }

    myProcess = ::GetCurrentProcessId();
}


/*
 * Class:     czerwony_kapturek_jbwheel_WheelTool
 * Method:    stopCaptureWheelMessages
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_czerwony_kapturek_jbwheel_WheelTool_stopCaptureWheelMessages(JNIEnv *, jobject)
{
    env->DeleteGlobalRef( jo );
    jo = NULL;
    env = NULL;
    vm = NULL;
    mid = 0;
    if (hhSend)
    {
        ::UnhookWindowsHookEx( hhSend );
        hhSend = 0;
    }
    if (hhPost)
    {
        ::UnhookWindowsHookEx( hhPost );
        hhPost = 0;
    }
}


JNIEXPORT void JNICALL
Java_czerwony_kapturek_jbwheel_WheelTool_setDebug(JNIEnv*, jobject, jboolean debug )
{
    trace("setDebug()");
    console = NULL;
    if (debug)
    {
        if (!tmpConsole)
        {
            AllocConsole();
            tmpConsole = GetStdHandle( STD_ERROR_HANDLE );
        }
        console = tmpConsole;
    }
}


//print debug messages to the console
void
trace( char* msg )
{
    if ( console )
    {
        printf("\njbWheel.dll: ");
        printf( msg );
        printf("\n");
    }
}



